/*--------------------------------------------------------------------

SLOW DOWN CPU - version 3.0

Copyright 2017-2019 Marcello Zaniboni

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 2.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

----------------------------------------------------------------------

Under Linux, compile with GCC using the foillowing command:
gcc -O3 -static -s -pthread slowdowncpu.c -o slowdowncpu
Under Windows, compile using Orwell Dev-C++ with TDM-GCC
(download: https://orwelldevcpp.blogspot.com/)

--------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <time.h>
#include <limits.h>
#include <signal.h>
#include <unistd.h>
#include <assert.h>

/* buffer length for user input */
#define INPUT_LEN 60
/* limit for the thread number the user can create */
#define MAX_THREADS 64
/* for dev use only */
#define DEBUG_PRINTS 0

/* thread data for statistics */
struct thread_data {
   int thread_id;    /* set by the main thread */
   long long t0;     /* start time */
   int exit_signal;  /* set by the main thread */
   unsigned long long int ops_counter;
};

struct thread_data thread_data_array[MAX_THREADS];

void print_welcome(void) {
	printf("\n");
	printf("---------------------------\n");
	printf("Slow Down CPU - version 3.0\n");
	printf("---------------------------\n");
	printf("   by Marcello Zaniboni\n");
	printf("marcello.zaniboni@gmail.com\n");
	printf("   mzaniboni@hotmail.com\n\n");
	printf("Every thread you create will probably fill one core of your CPU with\n");
	printf("arithmetic operations. Depending on your OS, if you want to slow down\n");
	printf("the CPU even more, it could also be convenient to run multiple\n");
	printf("instances of this software at the same time.\n");
}

void print_license(void) {
	printf("Software version: 3.0, September 2019\n");
	printf("check for upadtes here: https://www.marcellozaniboni.net/tinytools\n\n");
	printf(" This program is free software; you can redistribute it and/or modify\n");
	printf(" it under the terms of the GNU General Public License as published by\n");
	printf(" the Free Software Foundation, version 2 - http://www.gnu.org/\n");
}

int read_user_choice(void) {
	char *success;
	char choice[INPUT_LEN + 4];
	success = fgets(choice, INPUT_LEN, stdin);
	return success ? atoi(choice) : 0;
}

void *cpu_thread(void *arg) {
	struct thread_data *data;
	int a, b;
	float c;
	unsigned long int d, e;

	data = (struct thread_data *)arg;
	data->t0 = (long long)time(NULL);
	data->ops_counter = 0ULL;
	printf("- thread #%d started\n", (1 + data->thread_id));
	srand((unsigned int)time(0));
	while (!data->exit_signal) {
		a = rand();
		b = rand();
		c = (float)a / b;
		d = (unsigned long int)a + (unsigned long int)b;
		e = (unsigned long int)a * (unsigned long int)b;
		if (DEBUG_PRINTS)
				printf("t=%d n=%llu a=%d b=%d c=%f d=%lu e=%lu\n",
					(1 + data->thread_id),
					data->ops_counter,
					a, b, c, d, e);

		data->ops_counter++;
		if (data->ops_counter == (ULLONG_MAX - 1)) {
			data->t0 = (long long)time(NULL);
			data->ops_counter = 0ULL;
		}
	}
	printf("- thread #%d ended\n", (1 + data->thread_id));
	pthread_exit(NULL);
}

void print_thread_stats(struct thread_data *data) {
	long long curr_time;
	unsigned long long int dt;
	curr_time = (long long)time(NULL);
	dt = (unsigned long long int)(curr_time - data->t0);

	printf("thread #%d\n", (1 + data->thread_id));
	//printf("  loops:        %llu\n", data->ops_counter);
	printf("  sample time:  %lld seconds\n", dt);
	printf("  loops/second: %llu\n\n", dt > 0ULL ? (data->ops_counter / dt) : 0ULL);
}

int main(int argc, char *argv[]) {
	int choice, result;
	pthread_t mythreads[MAX_THREADS];
	int active_threads = 0;

	print_welcome();
	while (1) {
		sleep(1);
		if (active_threads == 0) {
			printf("\nStart a thread to load a CPU core.\n\n");
		} else {
			printf("\nActive threads: %d\n\n", active_threads);
		}
		printf("1. start a new thread\n");
		printf("2. delete a thread\n");
		printf("3. stats\n");
		printf("4. info and software license\n");
		printf("5. exit\n\n? ");
		choice = read_user_choice();
		printf("\n");
		if (choice == 1) {
			if (active_threads < MAX_THREADS) {
				thread_data_array[active_threads].thread_id = active_threads;
				thread_data_array[active_threads].exit_signal = 0;
				result = pthread_create(&mythreads[active_threads], NULL, cpu_thread, (void *) &thread_data_array[active_threads]);
				assert(!result);
				active_threads++;
			} else {
				printf("Error: too many threads running.\n");
			}
		} else if (choice == 2) {
			if (active_threads > 0) {
				active_threads--;
				thread_data_array[active_threads].exit_signal = 1;
			} else {
				printf("No threads running, nothing to delete.\n");
			}
		} else if (choice == 3) {
			if (active_threads > 0) {
				int i;
				for (i = 0; i < active_threads; i++) {
					print_thread_stats(&thread_data_array[i]);
				}
			} else {
				printf("No threads running, no stats.\n");
			}
		} else if (choice == 4) {
			print_license();
		} else if (choice == 5) {
			int i;
			for (i = 0; i < active_threads; i++)
				thread_data_array[i].exit_signal = 1;
			break;
		} else {
			printf("Invalid choice, please retry.\n");
		}
	}
	pthread_exit(NULL);

	return EXIT_SUCCESS;
}
